/* $Id: signal.c,v 1.1 2001/03/03 13:34:55 mikpe Exp $
 * signal.c
 *
 * This test program illustrates how performance counter overflow
 * can be caught and sent to the process as a user-specified signal.
 *
 * Limitations:
 * - Requires a kernel with UP-APIC support. Currently, this means
 *   a 2.4-ac kernel.
 * - The kernel driver doesn't yet identify which counter overflowed.
 *   Instead, all i-mode counters are stopped and it is the user's
 *   responsibility to resume them.
 * - Requires an Intel P6 or AMD K7 CPU.
 *
 * Copyright (C) 2001  Mikael Pettersson
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <signal.h>
#include <asm/sigcontext.h>
#include "linux/perfctr.h"
#define PAGE_SIZE	4096

static int fd;
static volatile const struct vperfctr_state *kstate;
static struct vperfctr_icontrol icontrol;
static unsigned cpu_type;

static void do_open(void)
{
    struct perfctr_info info;

    fd = open("/proc/self/perfctr", O_RDONLY);
    if( fd < 0 ) {
	perror("open");
	exit(1);
    }
    if( ioctl(fd, PERFCTR_INFO, &info) != 0 ) {
	perror("perfctr_info");
	exit(1);
    }
    cpu_type = info.cpu_type;
    if( !(info.cpu_features & PERFCTR_FEATURE_PCINT) )
	printf("PCINT not supported -- expect failure in icontrol()\n");
    kstate = mmap(NULL, PAGE_SIZE, PROT_READ, MAP_SHARED, fd, 0);
    if( kstate == MAP_FAILED ) {
	perror("mmap");
	exit(1);
    }
}

static void on_sigio(int sig, siginfo_t *si, void *puc)
{
    if( sig != SIGIO || si->si_code != 0x4711 ) {
	printf(__FUNCTION__ ": unexpected sig %d, code %#x\n", sig, si->si_code);
	return;
    }
    if( kstate->status <= 0 ) {
	/*
	 * My theory is that this happens if a perfctr overflowed
	 * at the very instruction for the ioctl(VPERFCTR_STOP) call.
	 * Signal delivery is delayed until the kernel returns to
	 * user-space, at which time VPERFCTR_STOP will already
	 * have marked the vperfctr as stopped. In this case, we
	 * cannot and must not attempt to IRESUME it.
	 * This can be triggered by counting e.g. BRANCHES and setting
	 * the overflow limit ridiculously low.
	 */
	printf(__FUNCTION__ ": unexpected overflow at pc %#lx (status %d)\n",
	       kstate->pc, kstate->status);
	return;
    }
    printf(__FUNCTION__ ": perfctr overflow at pc %#lx\n", kstate->pc);
    if( ioctl(fd, VPERFCTR_IRESUME, &icontrol.ctrs) < 0 ) {
	perror("iresume");
	abort();
    }
}

static void do_sigaction(void)
{
    struct sigaction sa;
    memset(&sa, 0, sizeof sa);
    sa.sa_sigaction = on_sigio;
    sa.sa_flags = SA_SIGINFO;
    if( sigaction(SIGIO, &sa, NULL) < 0 ) {
	perror("sigaction");
	exit(1);
    }
}

static void do_icontrol(void)
{
    unsigned evntsel;

    memset(&icontrol, 0, sizeof icontrol);
    switch( cpu_type ) {
      case PERFCTR_X86_INTEL_P6:
      case PERFCTR_X86_INTEL_PII:
      case PERFCTR_X86_INTEL_PIII:
	/* FLOPS, USR, ENable, INT */
	evntsel = 0xC1 | (1 << 16) | (1 << 22) | (1 << 20);
	break;
      case PERFCTR_X86_AMD_K7:
	/* K7 can't count FLOPS. Count RETIRED_TAKEN_BRANCHES instead. */
	evntsel = 0xC4 | (1 << 16) | (1 << 22) | (1 << 20);
	break;
      default:
	printf(__FUNCTION__ ": unsupported cpu type %u\n", cpu_type);
	exit(1);
    }	
    icontrol.control.evntsel[0] = evntsel;
    icontrol.ctrs.ctr[1] = -25; /* interrupt after 25 occurrences */
    icontrol.si_signo = SIGIO;
    icontrol.si_code = 0x4711;
    if( ioctl(fd, VPERFCTR_ICONTROL, &icontrol) < 0 ) {
	perror("icontrol");
	exit(1);
    }
}

static void do_stop(void)
{
    struct sigaction sa;

    if( ioctl(fd, VPERFCTR_STOP, NULL) )
	perror("stop");
    memset(&sa, 0, sizeof sa);
    sa.sa_handler = SIG_DFL;
    if( sigaction(SIGIO, &sa, NULL) < 0 ) {
	perror("sigaction");
	exit(1);
    }
}

#define N 150
static double v[N], w[N];
static double it;

static void do_dotprod(void)
{
    int i;
    double sum;

    sum = 0.0;
    for(i = 0; i < N; ++i)
	sum += v[i] * w[i];
    it = sum;
}

int main(void)
{
    do_sigaction();
    do_open();
    do_icontrol();
    do_dotprod();
    do_stop();
    return 0;
}
